-- TODOs:
--   serialization and unserialization
--   emissionTime, spawnBurst should be reasonable to support
--   emitter meshes would be easy to support

--[[
And here's where the modding steps into uncharted, deeply stupid waters.
I'm pretty sure Grimrock's native particle code handles emitting/killing
particles by modifying the mesh data. Nice and easy.
But there's no way to efficiently do that with just the game's Lua
interface (you have to get all your mesh data in Lua then copy it all
to a Mesh then turn it into a RenderableMesh and oh god it's so slow).

I'm not in the mood to hack native code or otherwise mess directly
with memory to the extent necessary to change that. Similar hacking
would be necessary to, like, make the game load my custom shader in
place of ParticleSystem.vsh or whatever.

Enter the totally-static-mesh particle system. Its features include:
- Branching in the vertex shader.
- Needing lots of different vertex shaders.
- Full support of ParticleEntity:start()/stop()-style control is totally impractical.
- Utterly incapable of usefully supporting "objectSpace = false".

]]


ZhandulParticleComponent = class(Component)

local particlePlaceholderMaterial = Material.create("particle_placeholder")

local zhandulEmitterDefs = {}

local pointsMeshCache = {}

function createTestMesh()
	local vertexArray = {}
	for i=1,12 do
		vertexArray[i] = i
	end
	local indices = {}
	indices[1] = 0
	indices[2] = 1
	indices[3] = 2
	local mesh = Mesh.create(#vertexArray/3)
	mesh:setVertexArray(vertexArray)
	mesh:setIndices(indices)
	mesh:addSegment(particlePlaceholderMaterial,"triangles",0,1)
	
	testMesh = mesh
	testRMesh = RenderableMesh.create(mesh)
end
createTestMesh()

-- returns RenderableMesh
local function getPointsMesh(count, ignoreCache)
	if not ignoreCache and pointsMeshCache[count] then
		return pointsMeshCache[count]
	end
	
	-- Each point gets a random x, y, z position between 0.0-1.0.
	-- Each point is a degenerate quad; four vertices at the same position
	-- (but see texcoords below).
	local vertexArray = {}
	for point=1,count do
		local x = math.random()
		local y = math.random()
		local z = math.random()
		local verti = point*12-11
		for i=1,4 do
			vertexArray[verti] = x
			vertexArray[verti+1] = y
			vertexArray[verti+2] = z
			verti = verti+3
		end
	end
	
	-- Texcoords are used the same way as in vanilla Grimrock particles
	-- (they store which corner of the quad the vertex is, and the vertex
	-- shader expands them later).
	local texcoordArray = {}
	for point=1,count do
		-- 4 vertices per point, 2 texcoords per vertex
		-- yes there are faster ways to fill this table...
		local ti = point*8-7
		texcoordArray[ti] = 0 -- NW
		texcoordArray[ti+1] = 1
		texcoordArray[ti+2] = 1 -- NE
		texcoordArray[ti+3] = 1
		texcoordArray[ti+4] = 0 -- SW
		texcoordArray[ti+5] = 0
		texcoordArray[ti+6] = 1 -- SE
		texcoordArray[ti+7] = 0
	end

	local indices = {}
	for point=1,count do
		local ii = point*6-5
		local vert1 = (point-1)*4
		indices[ii] = vert1 -- NW-NE-SW triangle
		indices[ii+1] = vert1+1
		indices[ii+2] = vert1+2
		indices[ii+3] = vert1+3 -- NE-SW-SE triangle
		indices[ii+4] = vert1+2
		indices[ii+5] = vert1+1
	end
	
	-- These arrays store random numbers.
	-- These are used for properties that should appear random per-particle.
	-- The tangent array cannot be used for this (just trust me here).
	
	-- Room for more random numbers:
	-- bitangents?
	-- 
	
	-- 8 bit ints
	local colors = {}
	for point=1,count do
		local x = math.random(-128,127)
		local y = math.random(-128,127)
		local z = math.random(-128,127)
		local w = math.random(-128,127)
		local verti = point*16-15
		for i=1,4 do
			colors[verti] = x
			colors[verti+1] = y
			colors[verti+2] = z
			colors[verti+3] = w
			verti = verti+4
		end
	end
	
	-- these get mapped from 0.0-1.0 to integer 0-255 before they make it to
	-- the shader, so our resolution here is just as bad as the color array
	local normalsButNotReally = {}
	for point=1,count do
		local x = math.random()*2-1
		local y = math.random()*2-1
		local z = math.random()*2-1
		local verti = point*12-11
		for i=1,4 do
			normalsButNotReally[verti] = x
			normalsButNotReally[verti+1] = y
			normalsButNotReally[verti+2] = z
			verti = verti+3
		end
	end
	
	

	local mesh = Mesh.create(#vertexArray/3)
	mesh:setVertexArray(vertexArray)
	mesh:setTexcoordArray(texcoordArray)
	mesh:setNormalArray(normalsButNotReally)

	mesh:setIndices(indices)
	mesh:setColorArray(colors)
	mesh:addSegment(particlePlaceholderMaterial,"triangles",0,#indices/3)
	
	mesh = RenderableMesh.create(mesh)
	if not ignoreCache then pointsMeshCache[count] = mesh end
	
	return mesh
end

function defineZhandulEmitter(def)
	-- what's a validation
	def = table.deepcopy(def)
	
	-- need either maxParticles or emissionRate; probably shouldn't have both.
	local maxParticles = math.ceil(def.maxParticles or (def.emissionRate * def.lifetime[2]))
	def.maxParticles = maxParticles
	if not def.emissionRate then
		def.dieTime = def.lifetime[2]
	end
	
	-- length of time before the emitter needs to start reusing quads
	-- not meaningful for systems with no emission rate
	def.emitterCycleTime = maxParticles/(def.emissionRate or 1)
	

	local boxMin = vec(def.boxMin[1], def.boxMin[2], def.boxMin[3])
	local boxMax = vec(def.boxMax[1], def.boxMax[2], def.boxMax[3])
	--[[
	local maxTime = def.lifetime[2]
	local maxEndVelocityVector = vec(gravity[1] > 0 and maxVelocity or minVelocity
					gravity[2] > 0 and maxVelocity or minVelocity,
					gravity[3] > 0 and maxVelocity or minVelocity)
	local maxDistance = gravity * (maxTime/airResistance) + (1/airResistance) * (maxVelocityVector - gravity*(1/airResistance)) * (1 - math.exp(-airResistance*maxTime))]]
	
	-- huge bounding box for testing
	-- TODO: compute a more reasonable one
	def.boundBox = Box.create((boxMin+boxMax)*0.5, vec(64,64,64))
	
	-- generate mesh in advance
	getPointsMesh(maxParticles)
	
	zhandulEmitterDefs[def.name] = def
end

function ZhandulParticleComponent.emitterDefined(name)
	return zhandulEmitterDefs[name] ~= nil
end

function ZhandulParticleComponent:init(go)
	self:super(go)
	self.node = self.go:createChildNode()
	self.timeScale = 1
end

function ZhandulParticleComponent:destroy()
	self.go.map.scene:removeNode(self.node, true)
	self.node = nil
	if self.materialInstance then self.materialInstance:dispose() end
end

function ZhandulParticleComponent:setEmitter(emitter)
	if emitter == self.emitter then return end
	local emitterDef = zhandulEmitterDefs[emitter]
	if not emitterDef then console:warn("no such zhandul emitter: "..emitter) return end
	
	self.emitter = emitter
	local me = MeshEntity.create(getPointsMesh(emitterDef.maxParticles))
	if self.sortOffset then me:setSortOffset(self.sortOffset) end
	
	self:initMaterialInstance(emitterDef)
	
	me:setMaterial(self.materialInstance)
	me:setManualBounds(emitterDef.boundBox, false) -- idk what the second parameter does lol
	self.node:setRenderEntity(me)
	
	self.meshEntity = me
	
	self:restart()
end

function ZhandulParticleComponent:initMaterialInstance(def)
	if self.materialInstance then self.materialInstance:dispose() end
	local material = Material.create("zhandulEmitterInstance")
	
	material:setLighting(false)
	material:setAmbientOcclusion(false)
	material:setCastShadow(false)
	material:setAlphaTest(false)
	material:setBlendMode(def.blendMode)
	material:setTextureAddressMode("Wrap")
	material:setDepthBias(0)
	material:setShader(ShaderEx.getShader(def.shader))
	
	local boxMin = vec(def.boxMin[1], def.boxMin[2], def.boxMin[3])
	local boxMax = vec(def.boxMax[1], def.boxMax[2], def.boxMax[3])
	material:setParam("g_particleBoxMin", boxMin, 3)
	material:setParam("g_particleBoxMax", boxMax, 3)
	material:setParam("g_particleLifetime", vec(def.lifetime[1], def.lifetime[2], def.emitterCycleTime), 3)
	local color = def.color0
	material:setParam("g_particleColor", vec(color[1], color[2], color[3], color[4]), 4)
	local gravity = def.gravity and vec(def.gravity[1], def.gravity[2], def.gravity[3]) or vec()
	material:setParam("g_particleGravity", gravity, 3)
	local minVelocity = def.velocity[1]
	local maxVelocity = def.velocity[2]
	if minVelocity > maxVelocity then minVelocity, maxVelocity = maxVelocity, minVelocity end
	material:setParam("g_particleMovementParams", vec(minVelocity, maxVelocity, math.rad(def.sprayAngle[1]), math.rad(def.sprayAngle[2])), 4)
	local airResistance = def.airResistance or 0.0001
	material:setParam("g_particleAirResistance", airResistance)
	material:setParam("g_particleRotationParams", vec(def.rotationSpeed or 0, def.randomInitialRotation == false and 0 or math.pi*2), 2);
	material:setParam("g_particleFadeParams", vec(1/def.fadeIn, 1/def.fadeOut, def.fadeOut), 3)
	--material:setParam("g_particleAnimationParams", vec(def.frameRate or 0, def.frameCount or 0, def.frameColumnCount or 0, def.frameSizePct or 0), 4)
	material:setParam("g_particleDepthBias", def.depthBias or 0)
	
	material:setParam("g_particleTime",0)
	material:setParam("g_timeSinceStop",0)
	
	-- custom shader parameters?
	if def.shaderParams then
		for param,value in pairs(def.shaderParams) do
			param = "g_"..param
			if type(value) == "table" then
				material:setParam(param, value[1], value[2])
			else
				material:setParam(param, value)
			end
		end
	end
	
	self.materialInstance = material
end

function ZhandulParticleComponent:setParam(name, param, length)
	if self.materialInstance then
		self.materialInstance:setParam(name, param, length)
	end
end

function ZhandulParticleComponent:getEmitter()
	return self.emitter;
end

function ZhandulParticleComponent:setSortOffset(sortOffset)
	self.sortOffset = sortOffset;
	local me = self.node:getRenderEntity()
	if me then me:setSortOffset(sortOffset) end
end

function ZhandulParticleComponent:getSortOffset()
	return self.sortOffset;
end

function ZhandulParticleComponent:enable()
	if not self.enabled then
		Component.enable(self)
		self.meshEntity:setHidden(false)
	end
end

function ZhandulParticleComponent:disable()
	if self.enabled then
		Component.disable(self)
		self.meshEntity:setHidden(true)
	end
end

function ZhandulParticleComponent:update()
	if self.enabled and self.emitter then
		local dt = Time.deltaTime
		self.particleTime = self.particleTime+dt*self.timeScale
		local timeSinceStop = self.timeSinceStop
		if timeSinceStop then
			timeSinceStop = timeSinceStop+dt*self.timeScale
			self.timeSinceStop = timeSinceStop
		end
		
		if self:isDone() then
			if self.destroyObject then
				self.go:destroy()
				return
			end
			if self.destroySelf then
				self.go:removeComponent(self)
				return
			end
			if self.disableSelf then
				self:disable()
				return
			end
			self.meshEntity:setHidden(true) -- important for negative timescale
		elseif self.enabled then
			self.meshEntity:setHidden(false)
		end
		
		self.materialInstance:setParam("g_particleTime",self.particleTime)
		self.materialInstance:setParam("g_timeSinceStop",timeSinceStop or 0)
	end
end

function ZhandulParticleComponent:setDestroySelf(d)
	self.destroySelf = d
end
function ZhandulParticleComponent:getDestroySelf()
	return self.destroySelf
end

function ZhandulParticleComponent:setDestroyObject(d)
	self.destroyObject = d
end
function ZhandulParticleComponent:getDestroyObject()
	return self.destroyObject
end

function ZhandulParticleComponent:setTimeScale(t)
	self.timeScale = t
end
function ZhandulParticleComponent:getTimeScale()
	return self.timeScale
end

function ZhandulParticleComponent:setTime(t)
	self.particleTime = t
end
function ZhandulParticleComponent:getTime()
	return self.particleTime
end

function ZhandulParticleComponent:isDone()
	local emitter = self.emitter
	if not emitter then return true end
	
	if self.timeScale < 0 then return self.particleTime < 0 end
	
	local def = zhandulEmitterDefs[emitter]
	if def.dieTime then return self.particleTime >= def.dieTime end
	
	local timeSinceStop = self.timeSinceStop
	return timeSinceStop and timeSinceStop >= def.emitterCycleTime or false
end

function ZhandulParticleComponent:stop()
	if not self.timeSinceStop then self.timeSinceStop = 0 end
end
function ZhandulParticleComponent:restart()
	self.particleTime = 0
	self.timeSinceStop = nil
end
