-- Quaternion stuff so we can interpolate rotations in any useful way at all
-- I throw four numbers around instead of using a table, for performance
local math = math
local math_acos = math.acos
local math_cos = math.cos
local math_sin = math.sin

quat_normalize = function(x,y,z,w)
	local mult = 1/(x*x+y*y+z*z+w*w)^0.5
	return x*mult,y*mult,z*mult,w*mult
end

quat_dot = function(ax,ay,az,aw,bx,by,bz,bw)
	return ax*bx+ay*by+az*bz+aw*bw
end

-- Assumes the inputs are already normalized! Beware!
-- Could be optimized a lot further
quat_slerpFast = function(ax,ay,az,aw,bx,by,bz,bw,t)
	local dot = ax*bx+ay*by+az*bz+aw*bw
	-- can't slerp if the dot product is 1 or -1 (or sufficiently close, left a lot of margin here though)
	if dot*dot > 0.9998 then
		-- fall back to lerp
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,ay*amult+by*t,ay*amult+by*t
	end
	-- in the event of a handedness mismatch, invert the first
	-- quaternion
	if dot < 0 then
		ax = -ax
		ay = -ay
		az = -az
		aw = -aw
		dot = -dot
	end
	local angle = math_acos(dot)*t
	local cx = bx - ax*dot
	local cy = by - ay*dot
	local cz = bz - az*dot
	local cw = bw - aw*dot
	
	-- normalize
	local mult = 1/(cx*cx+cy*cy+cz*cz+cw*cw)^0.5
	cx = cx*mult
	cy = cy*mult
	cz = cz*mult
	cw = cw*mult
	
	local cos = math_cos(angle)
	local sin = math_sin(angle)
	return ax*cos+cx*sin,ay*cos+cy*sin,az*cos+cz*sin,aw*cos+cw*sin
end

-- Assumes the inputs are already normalized! Beware!
quat_lerpFast = function(ax,ay,az,aw,bx,by,bz,bw,t)
	local amult = 1-t
	return ax*amult+bx*t,ay*amult+by*t,ay*amult+by*t,ay*amult+by*t
end

-- Quaternion from Euler angles (radians)
quat_fromAngles = function(yaw,pitch,roll)
	local sinY = math_sin(yaw*0.5)
	local sinP = math_sin(pitch*0.5)
	local sinR = math_sin(roll*0.5)
	local cosY = math_cos(yaw*0.5)
	local cosP = math_cos(pitch*0.5)
	local cosR = math_cos(roll*0.5)
	return cosY*cosP*sinR-sinY*sinP*cosR,
		sinY*cosP*sinR+cosY*sinP*cosR,
		sinY*cosP*cosR-cosY*sinP*sinR,
		cosY*cosP*cosR+sinY*sinP*sinR
end

-- Quaternion from rotation matrix
-- This is written to provide the same result as Grimrock quat.fromMatrix()
quat_fromMatrix = function(mat)
	local matX = mat.x
	local matY = mat.y
	local matZ = mat.z
	
	local matX1 = matX[1]
	local matY2 = matY[2]
	local matZ3 = matZ[3]

	local trace = matX1+matY2+matZ3+1
	
	if trace > 0.0001 then
		local mult = 0.5/trace^0.5
		return (matY[3]-matZ[2])*mult,(matZ[1]-matX[3])*mult,(matX[2]-matY[1])*mult,0.25/mult
	elseif matX1 > matY2 and matX1 > matZ3 then
		local mult = 2*(1+matX1-matY2-matZ3)^0.5
		return 0.25*mult,(matY[1]+matX[2])/mult,(matZ[1]+matX[3])/mult,(matZ[2]-matY[3])/mult
	elseif matY2 > matZ3 then
		local mult = 2*(1+matY2-matX1-matZ3)^0.5
		return (matY[1]+matX[2])/mult,0.25*mult,(matZ[2]+matY[3])/mult,(matZ[1]-matX[3])/mult
	else
		local mult = 2*(1+matZ3-matX1-matY2)^0.5
		return (matZ[1]+matX[3])/mult,(matZ[2]+matY[3])/mult,0.25*mult,(matY[1]-matX[2])/mult
	end
end

-- and this is written to provide the same result as Grimrock quat.toMatrix()
-- when called without the destination argument
quat_toMatrix = function(x,y,z,w,destination)	
	local rval = destination or mat.identity()
	local rvalX = rval.x
	local rvalY = rval.y
	local rvalZ = rval.z
	
	rvalX[1] = 1-(y*2*y+z*2*z)
	rvalX[2] = y*2*x+z*2*w
	rvalX[3] = z*2*x-y*2*w
	rvalY[1] = y*2*x-z*2*w
	rvalY[2] = 1-(x*2*x+z*2*z)
	rvalY[3] = z*2*y+x*2*w
	rvalZ[1] = z*2*x+y*2*w
	rvalZ[2] = z*2*y-x*2*w
	rvalZ[3] = 1-(x*2*x+y*2*y)

	return rval
end
