-- Moves/rotates objects between a start and end point, with callbacks.
local function getNormalizedTime(data)
	return data.elapsed/data.time
end

local function callTransformationCallback(callback,...)
	if type(callback) == "table" then
		local f = callback[1]
		f(...,unpack(callback,2))
	else
		callback(...)
	end
end

local quatSlerpFast = quat_slerpFast
local quatToMatrix = quat_toMatrix
local quatFromAngles = quat_fromAngles
local quatFromMatrix = quat_fromMatrix
local quatNormalize = quat_normalize

local lostInTransformation
lostInTransformation = function(id,data)
	if data.components then
		for componentName,data2 in pairs(data.components) do
			lostInTransformation(id.."."..componentName,data2)
		end
	end
	local type = type
	if data.translation and type(data.translation) == "table"
		and type(data.translation.callback) == "table"
		and data.translation.callback.onLost then
			callTransformationCallback(data.translation.callback.onLost,id,getNormalizedTime(data.translation))
	end
	if data.transformation and type(data.transformation) == "table"
		and type(data.transformation.callback) == "table"
		and data.transformation.callback.onLost then
			callTransformationCallback(data.transformation.callback.onLost,id,getNormalizedTime(data.transformation))
	end
end

-- Table of moving GameObjects and corresponding data.
-- A misnomer in this particular dungeon, where it in fact isn't saved at all.
savedData = {tweenObjects = {}}

local function updateTranslation(gobj,data,dtime)
	local isComponent = gobj.go ~= nil
	
	local oldtime = getNormalizedTime(data)
	data.elapsed = math.min(data.time,data.elapsed+dtime)
	local newtime = getNormalizedTime(data)

	local ox,oy,oe = gobj.x, gobj.y, gobj.elevation
	if isComponent then
		gobj:setOffset(vec(data.interpolator(data[1],data[2],data[3],data[4],data[5],data[6],newtime)))
	else
		gobj:setWorldPosition(data.interpolator(data[1],data[2],data[3],data[4],data[5],data[6],newtime))
	end

	if data.callback then
		local type = type
		if type(data.callback) == "function" then
			if newtime == 1 then callTransformationCallback(data.callback,gobj,newtime) end
		elseif type(data.callback) == "table" then
			if data.callback[1] and data.callback[2] then -- Assume single callback with extra args
				callTransformationCallback(data.callback,gobj,newtime)
			else -- Multiple callbacks
				if data.callback.onUpdate then
					callTransformationCallback(data.callback.onUpdate,gobj,newtime)
				end
				if not isComponent and data.callback.onPositionChanged and (ox ~= gobj.x or oy ~= gobj.y or oe ~= gobj.elevation) then
					callTransformationCallback(data.callback.onPositionChanged,gobj,newtime,ox,oy,oe)
				end
				for k,v in pairs(data.callback) do
					if type(k) == "number" and oldtime < k and newtime >= k then
						callTransformationCallback(v,gobj,newtime)
					end
				end
			end
		end
	end

	if newtime == 1 then
		local tweenObjects = savedData.tweenObjects
		if isComponent then
			tweenObjects[gobj.go.id].components[gobj:getName()].translation = nil
		else
			tweenObjects[gobj.id].translation = nil
		end
	end
end

local function snapTranslation(gobj,data)
	updateTranslation(gobj,data.translation,math.huge)
end

local function updateRotation(gobj,data,dtime)
	local gobjId = gobj.id

	local oldtime = getNormalizedTime(data)
	data.elapsed = math.min(data.time,data.elapsed+dtime)
	local newtime = getNormalizedTime(data)

	gobj:setWorldRotation(quatToMatrix(data.interpolator(data[1],data[2],data[3],data[4],data[5],data[6],data[7],data[8],newtime)))

	if data.callback then
		local type = type
		if type(data.callback) == "function" then
			if newtime == 1 then callTransformationCallback(data.callback,gobj,newtime) end
		elseif type(data.callback) == "table" then
			if data.callback[1] and data.callback[2] then -- Assume single callback with extra args
				callTransformationCallback(data.callback,gobj,newtime)
			else -- Multiple callbacks
				if data.callback.onUpdate then
					callTransformationCallback(data.callback.onUpdate,gobj,newtime)
				end
				for k,v in pairs(data.callback) do
					if type(k) == "number" and oldtime < k and newtime >= k then
						callTransformationCallback(v,gobj,newtime)
					end
				end
			end
		end
	end

	if newtime == 1 then
		savedData.tweenObjects[gobjId].rotation = nil
	end
end

local function snapRotation(gobj)
	updateRotation(gobj,savedData.tweenObjects[gobj.id].rotation,math.huge)
end
tween_snapRotation = snapRotation

function _tween_update(dtime)
	local tweenObjects = savedData.tweenObjects
	for gobjId,data in pairs(tweenObjects) do
		local gobj = ScriptComponent.baseEnv.findEntity(gobjId)
		if not gobj then
			lostInTransformation(gobjId,data)
			tweenObjects[gobjId] = nil
			return
		end
		if data.components then
			local notEmpty
			for componentName,data2 in pairs(data.components) do
				if data2.translation then
					updateTranslation(gobj:getComponent(componentName),data2.translation,dtime)
				end
				if data2.translation then
					notEmpty = true
				else
					data.components[componentName] = nil
				end
			end
			if not notEmpty then
				data.components = nil
			end
		end
		if data.rotation then
			updateRotation(gobj,data.rotation,dtime)
		end
		if data.translation then
			updateTranslation(gobj,data.translation,dtime)
		end
		
		if not (data.components or data.rotation or data.translation) then
			tweenObjects[gobjId] = nil
		end
	end
end

-- Transform an object over a period of time.
--
-- If you want to set the object's transform instantly, just use
-- GameObject:setWorldRotation() and GameObject:setWorldPosition() instead of this.
--
-- There are currently two functions for this:

-- rotateOverTime: For rotating an object only.
-- newRotation can be a vec or matrix. A vec is interpreted as (x,y,z) rotation
-- in degrees, e.g. vec(45,90,0) to rotate 45 degrees on the x axis and 90 degrees
-- on the y axis. Note that both are converted to quats, so the interpolation
-- will always be the same.
-- XXX: Warning: Using rotateOverTime will currently remove any non-rotation
-- component of the object's transformation matrix, so don't call it on an
-- object that is currently sheared or scaled.

-- translateOverTime: For translating objects. This changes the object's world
-- position instead of its transformation matrix, and can be used concurrently
-- with rotateOverTime.

-- gobj: Object to transform
-- newRotation/Position: Endpoint for transformation. (The start point
--   is the object's current transform.) Can be a temporary matrix.
-- time: Time in seconds that it takes for the transformation to complete.
-- interpolator: Interpolation function for intermediate transforms.
--   This function receives as arguments the original x,y,z,(w), the final
--   x,y,z,(w), and the normalized (0-1) elapsed time, in that order, and must
--   return x,y,z,(w). It is called every frame during movement/rotation.
--
-- callback: Function(s) to be called at the end of transformation or during it.
--   The first two arguments are always the gobj and the normalized time. For all
--   callbacks EXCEPT "onLost", the gobj is guaranteed not nil.
--   Because callback functions get stored in this ScriptComponent they should
--   be environment-agnostic!
--
--   callback can be any of the following:
--     nil
--
--     a function (will be called immediately after the end of the transformation)
--
--     an array where the first element is a function as above, and subsequent elements
--     are extra arguments. Example:
--     {function(gobj,t,snd) playSound(snd) end, "clunk"}
--
--     a table with any of the above as values. Keys can be one of several things:
--
--     1. Number keys are interpreted as normalized times; the values will be called
--     as above *after* that normalized time is reached. So passing
--     {[1]=doEndStuff} is equivalent to just passing doEndStuff.
--     Multiple time-based callbacks can be called on the same frame, should the
--     transformation pass through both of them on the same frame, but they will never
--     be called out of order.
--     A callback with a key of 1 will always be called after all other callbacks.
--
--     2. Special string keys:
--
--     "onUpdate": Called immediately after every frame of transformation and before
--     all other callbacks except onLost.
--
--     "onLost": Called when findEntity() fails to find the GameObject, meaning that
--     the object is not on any map - either it was destroyed, or it moved to an inventory.
--     Since the GameObject cannot be found, the first argument to an onLost callback will
--     always be nil. The second argument is still the normalized time. The third argument
--     is the id of the object that was lost.
--     After the onLost hook is called, the transformation will stop, since it no longer
--     has an object to transform. No other callbacks will be called.
--
--     "onPositionChanged": For translateOverTime only.
--     Called *after* the object's map position (x, y, or elevation)
--     changes. This function's first arguments are (gobj,time,oldX,oldY,oldElevation).
--     Do not write code that depends on onPositionChanged being called before or after
--     a < 1.0 time-based callback; whether it will be called before or after a time-based
--     callback with a non-1 key is _undefined_.
--     As stated above, a callback with a key of 1 will always be called after all other
--     callbacks, including onPositionChanged, so you CAN rely on that.
--
--     Example:
--     {
--       [0.5] = {function(gobj,t,snd) playSound(snd) end,"clunk"},
--       [1] = doEndStuff,
--       onPositionChanged = doPositionChangedStuff,
--       onUpdate = doUpdateStuff,
--     }
--
-- If the object is already being translated/rotated, that existing transform will be
-- "fast-forwarded" and finish immediately.
local TRANSLATION_INTERPOLATORS = {
	LINEAR = function(ax,ay,az,bx,by,bz,t)
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,az*amult+bz*t
	end,
	ATTACK = function(ax,ay,az,bx,by,bz,t) -- Fast at the start, slow at the end
		t = t^0.5
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,az*amult+bz*t
	end,
	ACCEL = function(ax,ay,az,bx,by,bz,t) -- Slow at the start, fast at the end
		t = t^3
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,az*amult+bz*t
	end,
	BOUNCEBACK = function(ax,ay,az,bx,by,bz,t) -- Exceed the target a bit and bounce back
		t = t^0.6+t^2-t^4
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,az*amult+bz*t
	end,
	BUTTON_DOWN = function(ax,ay,az,bx,by,bz,t) -- Exceed the target a bit and bounce back but gently
		t = t^0.6*1.2+t^2-t^3*1.2
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,az*amult+bz*t
	end,
	BUTTON_UP = function(ax,ay,az,bx,by,bz,t) -- Go backwards a bit then spring up
		t = t-t^0.6+t^1.2-t^1.6+t^2
		local amult = 1-t
		return ax*amult+bx*t,ay*amult+by*t,az*amult+bz*t
	end,
}
local ROTATION_INTERPOLATORS = {
	-- Use LINEAR only if you want to preserve the handedness of the start
	-- quaternion even when flipping handedness would result in a shorter
	-- path. It is unlikely that you want to do that.
	LINEAR = quatLerpFast,
	SPHERICAL = quatSlerpFast,
	SPHERICAL_ATTACK = function(ax,ay,az,aw,bx,by,bz,bw,t) -- Fast at the start, slow at the end
		t = t^0.5
		return quatSlerpFast(ax,ay,az,aw,bx,by,bz,bw,t)
	end,
	SPHERICAL_BOUNCEBACK = function(ax,ay,az,aw,bx,by,bz,bw,t) -- Exceed the target a bit and bounce back
		t = t^0.6+t^2-t^4
		return quatSlerpFast(ax,ay,az,aw,bx,by,bz,bw,t)
	end,
	SPHERICAL_ACCEL = function(ax,ay,az,aw,bx,by,bz,bw,t) -- Slow at the start, fast at the end
		t = t^4
		return quatSlerpFast(ax,ay,az,aw,bx,by,bz,bw,t)
	end,
	SPHERICAL_SMOOTHSTEP = function(ax,ay,az,aw,bx,by,bz,bw,t)
		t = t*t*(3-2*t)
		return quatSlerpFast(ax,ay,az,aw,bx,by,bz,bw,t)
	end,
	SPHERICAL_SMOOTHERSTEP = function(ax,ay,az,aw,bx,by,bz,bw,t)
		t = t*t*t*(t*(t*6-15)+10)
		return quatSlerpFast(ax,ay,az,aw,bx,by,bz,bw,t)
	end,
}
function tween_translateOverTime(gobj,newPosition,time,interpolator,callback)
	local type = type
	if not vec.isvec(newPosition) then
		Console.warn(string.format("translateOverTime: bad newPosition (expected vec, got %s)",type(newPosition)))
		return
	end

	if type(time) ~= "number" then
		Console.warn(string.format("translateOverTime: bad time (expected number, got %s)",type(time)))
		return
	end
	if time <= 0 or time ~= time then -- checks if time is NaN
		Console.warn("minlib.gobj.translateOverTime: time must be greater than 0")
		return
	end

	if type(interpolator) == "string" then interpolator = TRANSLATION_INTERPOLATORS[interpolator:upper()] end

	if type(interpolator) ~= "function" then
		Console.warn(string.format("minlib.gobj.translateOverTime: bad interpolator (expected function, got %s)",type(interpolator)))
		return
	end
	
	local dataTable
	local startPoint
	local tweenObjects = savedData.tweenObjects
	if gobj.go then
		local id = gobj.go.id
		if not tweenObjects[id] then tweenObjects[id] = {} end
		local gData = tweenObjects[id]
		if not gData.components then gData.components = {} end
		local name = gobj:getName()
		if not gData.components[name] then gData.components[name] = {} end
		if gData.components[name].translation then
			snapTranslation(gData.components[name])
		end
		dataTable = gData.components[name]
		startPoint = gobj:getOffset()
	else
		local id = gobj.id
		if not tweenObjects[id] then tweenObjects[id] = {} end
		dataTable = tweenObjects[id]
		if dataTable.translation then
			snapTranslation(gobj,dataTable)
		end
		startPoint = gobj:getWorldPosition()
	end
	dataTable.translation = {
		startPoint[1],
		startPoint[2],
		startPoint[3],
		newPosition[1],
		newPosition[2],
		newPosition[3],
		time = time,
		elapsed = 0,
		interpolator = interpolator,
		callback = callback,
	}
end
function tween_rotateOverTime(gobj,newRotation,time,interpolator,callback)
	local endX,endY,endZ,endW
	local type = type
	if type(newRotation) == "table" then
		if vec.isvec(newRotation) then
			-- XXX: may want to change this to quatFromMatrix(FUNCTIONS.mat_getRotationMatrix(...))
			-- to ensure consistency with quats obtained from worldRotations
			endX,endY,endZ,endW = quatFromAngles(math.rad(newRotation[1]),math.rad(newRotation[2]),math.rad(newRotation[3]))
		else
			endX,endY,endZ,endW = quatFromMatrix(newRotation)
		end
	else
		Console.warn(string.format("rotateOverTime: bad newRotation (expected vec or matrix, got %s)",type(newRotation)))
		return
	end
	if type(time) ~= "number" then
		Console.warn(string.format("rotateOverTime: bad time (expected number, got %s)",type(time)))
		return
	end
	if time <= 0 or time ~= time then -- checks if time is NaN
		Console.warn("rotateOverTime: time must be greater than 0")
		return
	end

	if type(interpolator) == "string" then interpolator = ROTATION_INTERPOLATORS[interpolator:upper()] end

	if type(interpolator) ~= "function" then
		Console.warn(string.format("rotateOverTime: bad interpolator (expected function, got %s)",type(interpolator)))
		return
	end

	local gobjId = gobj.id
	local tweenObjects = savedData.tweenObjects
	if not tweenObjects[gobjId] then tweenObjects[gobjId] = {} end
	local gData = tweenObjects[gobjId]
	if gData.rotation then
		snapRotation(gobj)
	end
	
	local startX,startY,startZ,startW = quatNormalize(quatFromMatrix(gobj:getWorldRotation()))
	endX,endY,endZ,endW = quatNormalize(endX,endY,endZ,endW)
	gData.rotation = {
		startX,
		startY,
		startZ,
		startW,
		endX,
		endY,
		endZ,
		endW,
		time = time,
		elapsed = 0,
		interpolator = interpolator,
		callback = callback,
	}
end
function tween_isMoving(gobj)
	return savedData.tweenObjects[gobj.id] ~= nil
end
