-- VERY un-fancy crap preprocessor parser, because doing it properly is a pain
-- and this is unlikely to be used by anyone except me.
-- Directives have to be at the beginning of a line, and it doesn't check the
-- context to see if they're actually part of a long string or something.
-- Only two directives are supported:
--
-- 1. #define does what you'd expect, mostly. Identifiers can only contain
--     CAPITAL LETTERS (in the hopes of preventing confusion with my actual Lua
--     identifiers), underscores, and digits, and cannot begin with a digit.
--     Replacement tokens are unrestricted but trailing spaces will be removed.
--
--     To be expanded when used in the source, this identifier must be bordered
--     by characters (or the beginning/end of the file) that are not valid for
--     identifiers. So if you #define EXAMPLE, then it will be expanded in the
--     strings "EXAMPLE", "EXAMPLE * 2", "print(EXAMPLE*2)", but will not be
--     recognized in "EXAMPLE2" or "bad_EXAMPLE".
--     This is for the sake of being able to use string.gsub() efficiently; Lua
--     patterns are quite limited compared to, say, regexes!
--
--     Function-like macros are not supported.
--
-- 2. #includeDefines [filename] copies all the defined identifiers from that
--     file (including the results of that file's #includeDefines).
--
-- So what is this actually used for? String/number/boolean constants. Easier to
-- maintain than magic numbers, and easier to use across files than table
-- fields, and MUCH more performant than table fields in most cases.
local preprocessIncludes = {}
local definesByFile = {}
local localDefinesByFile = {}
local includeChain = {}

local function _getIncludeChainString()
	if #includeChain == 0 then
		error("_getIncludeChainString() called outside _getDefines()")
	end
	local rval = "in "
	for i=#includeChain,1,-1 do
		rval = rval..includeChain[i]
		if i ~= 1 then
			rval = rval.."\nincluded from "
		end
	end
	return rval
end

local function _getDefines_error(text)
	local chainString = _getIncludeChainString()
	includeChain = {} -- error means the function won't return so we have to reset includeChain manually here
	preprocessIncludes = {}
	error(string.format("%s\n%s",text,chainString))
end

-- warning this function is awful. source can be nil but filename cannot
local function _getDefines(source, filename, includeLocal)
	includeChain[#includeChain+1] = filename
	local defs = definesByFile[filename]
	if includeLocal or not defs then
		defs = {}
		local localDefs = {}
		if not source then source = FileSystem.readFile(filename) end
		local index = 1
		local length = source:len()
		local definePattern = "^#define([^%s]*) ([^%s]+) ([^\n]+)"
		local includePattern = "^#includeDefines ([^\n]+)"
		local invalidPattern = "^(#[^\n]+)"
		while index < length do
			local defType,defName,defValue = source:match(definePattern,index)
			if defName then
				local isLocalDef = defType == "Local"
				if not isLocalDef and defType ~= "" then
					_getDefines_error(string.format("invalid #define identifier: %s\n%s",defName,_getIncludeChainString()))
				end
				
				if defName:match("[^%a%d_]") or defName:match("^%d") then
					_getDefines_error(string.format("invalid #define identifier: %s\n%s",defName,_getIncludeChainString()))
				end
			
				if defs[defName] or localDefs[defName] then
					console:warn(string.format("redefinition of %s\n%s",defName,_getIncludeChainString()))
				end
				
				-- remove trailing whitespace
				defValue = defValue:gsub("%s*$","")
				
				if includeLocal then localDefs[defName] = defValue end
				if not isLocalDef then defs[defName] = defValue end
			else
				local includeFilename = source:match(includePattern,index)
				
				if includeFilename then
					-- this section prevents infinite recursion
					if preprocessIncludes[includeFilename] then
						_getDefines_error(string.format("multiple #includeDefines of %s\n%s",includeFilename,_getIncludeChainString()))
					end
					preprocessIncludes[includeFilename] = true
					
					for defName,defValue in pairs(_getDefines(nil,includeFilename)) do
						if defs[defName] then
							console:warn(string.format("redefinition of %s\n%s",defName,_getIncludeChainString()))
						end
						defs[defName] = defValue
					end
					includeChain[#includeChain] = nil
				else
					local invalidDirective = source:match(invalidPattern)
					if invalidDirective then
						_getDefines_error(string.format("invalid preprocessor directive: %s\n%s",invalidDirective,chainString))						
					end
				end
			end
			index = source:find("\n#",index)
			if index then index = index + 1 else break end
		end
		definesByFile[filename] = defs
	end
	return defs
end
local function _preprocess(source, filename)
	local defs = _getDefines(source, filename, true)
	includeChain = {}
	preprocessIncludes = {}
	definesByFile = {}
	
	-- resolve #defined strings in values
	for defName,defValue in pairs(defs) do
		local prevValue
		repeat
			prevValue = defValue
			defValue = defValue:gsub("[%a%d_]*",defs)
		until prevValue == defValue
		defs[defName] = defValue
	end

	-- remove preprocessor directives from source
	source = source:gsub("\n#[^\n]+","\n") -- keep newline so that line numbers match
	source = source:gsub("^#[^\n]+","") -- directive on first line
	 
	-- finally actually expand the tokens!
	source = source:gsub("[%a%d_]*",defs)
	
	return source
end

function loadfile_withPreprocessor(filename)
	return loadstring_withPreprocessor(FileSystem.readFile(filename), filename)
end
function loadstring_withPreprocessor(str, filename)
	return loadstring(_preprocess(str, filename or str), filename)
end
